# -*- coding: utf-8 -*-
import socket
import sys
import time

import Cryptodome.PublicKey.RSA
import pexpect
import stem.control

import onionbalance.hs_v2.util
from test.functional.util import *


def test_master_descriptor_publication(tmpdir):
    """
    Functional test to run Onionbalance, publish a master descriptor and
    check that it can be retrieved from the DHT.
    """

    chutney_config = parse_chutney_enviroment()
    print(chutney_config)
    private_key = Cryptodome.PublicKey.RSA.generate(1024)
    master_onion_address = onionbalance.hs_v2.util.calc_onion_address(private_key)

    config_file_path = create_test_config_file_v2(
        tmppath=tmpdir,
        private_key=private_key,
        instances=chutney_config.get('instances', []),
    )
    assert config_file_path

    # Start an Onionbalance server and monitor for correct output with pexpect
    server = pexpect.spawnu("onionbalance",
                            args=[
                                '--hs-version', 'v2',
                                '-i', chutney_config.get('client_ip'),
                                '-p', str(chutney_config.get('control_port')),
                                '-c', config_file_path,
                                '-v', 'debug',
                                '--is-testnet'
                            ], logfile=sys.stdout, timeout=15)

    # Check for expected output from Onionbalance
    server.expect(u"Loaded the config file")
    server.expect(u"introduction point set has changed")
    server.expect(u"Published a descriptor", timeout=60)

    # Check Tor control port gave an uploaded event.

    server.expect(u"HS_DESC UPLOADED")
    # Eek, sleep to wait for descriptor upload to all replicas to finish
    time.sleep(10)

    # .. todo:: Also need to check and raise for any warnings or errors
    #           that are emitted

    # Try fetch and validate the descriptor with stem
    with stem.control.Controller.from_port(
        address=chutney_config.get('client_ip'),
        port=chutney_config.get('control_port')
    ) as controller:
        controller.authenticate()

        # get_hidden_service_descriptor() will raise exceptions if it
        # cannot find the descriptors
        master_descriptor = controller.get_hidden_service_descriptor(
            master_onion_address)
        master_ips = master_descriptor.introduction_points()

        # Try retrieve a descriptor for each instance
        for instance_address in chutney_config.get('instances'):
            instance_descriptor = controller.get_hidden_service_descriptor(
                instance_address.split(':')[0])
            instance_ips = instance_descriptor.introduction_points()

            # Check if all instance IPs were included in the master descriptor
            assert (set(ip.identifier for ip in instance_ips) ==
                    set(ip.identifier for ip in master_ips))

    # Check that the control socket was created
    socket_path = tmpdir.join('control')
    assert socket_path.check()

    # Connect to the control socket and check the output
    sock_client = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    sock_client.connect(str(socket_path))

    # Read the data from the status socket
    result = []
    while True:
        data = sock_client.recv(1024)
        if not data:
            break
        result.append(data.decode('utf-8'))
    result_data = ''.join(result)

    # Check each instance is in the output
    for instance_address in chutney_config.get('instances'):
        assert instance_address.split(':')[0] in result_data

    # Check all instances were online and all master descriptors uploaded
    assert master_onion_address in result_data
    assert '[offline]' not in result_data
    assert '[not uploaded]' not in result_data




# -*- coding: utf-8 -*-
"""
Provide status over Unix socket
Default path: /var/run/onionbalance/control
"""
import json
import threading
import socket
from socketserver import BaseRequestHandler, ThreadingMixIn, UnixStreamServer

from onionbalance.common import log
from onionbalance.common.status import BaseStatusSocket
from onionbalance.hs_v3.onionbalance import Onionbalance

logger = log.get_logger()


class StatusSocketHandlerMixin(object):
    def __init__(self, balance: Onionbalance):
        self.balance = balance

    def _outputString(self):
        time_format = "%Y-%m-%d %H:%M:%S"
        response = {}
        services_data = []
        for service in self.balance.services:
            instances_data = []
            for instance in service.instances:
                received = instance.descriptor.received_ts.strftime(time_format) \
                    if instance.descriptor else None
                if not instance.intro_set_modified_timestamp:
                    instances_data.append({'onionAddress': instance.onion_address + '.onion',
                                           'received': received})
                else:
                    instances_data.append({'onionAddress': instance.onion_address + '.onion',
                                           'introSetModified': instance.intro_set_modified_timestamp.strftime(time_format),
                                           'introPointsNum': len(instance.descriptor.intro_set),
                                           'descriptorReceived': received})
            if service.first_descriptor:
                last_attempt_first = service.first_descriptor.last_publish_attempt_ts.strftime(time_format)
            else:
                last_attempt_first = None
            if service.second_descriptor:
                last_attempt_second = service.second_descriptor.last_publish_attempt_ts.strftime(time_format)
            else:
                last_attempt_second = None

            services_data.append({'onionAddress': service.onion_address,
                                  'publishAttemptFirstDescriptor': last_attempt_first,
                                  'publishAttemptSecondDescriptor': last_attempt_second,
                                  'instances': instances_data})
        response['services'] = services_data
        return json.dumps(response, sort_keys=True)


class StatusSocketHandlerImpl(BaseRequestHandler, StatusSocketHandlerMixin):
    """
    Handler for new domain socket connections
    """

    def __init__(self, balance: Onionbalance, *args, **kwargs):
        StatusSocketHandlerMixin.__init__(self, balance)
        BaseRequestHandler.__init__(self, *args, **kwargs)

    def handle(self):
        """
        Prepare and output the status summary when a connection is received
        """
        self.request.sendall(self._outputString().encode('utf-8'))


def create_status_socket_handler(balance: Onionbalance):
    def StatusSocketHandler(*args, **kwargs):
        """
        Fake constructor for StatusSocketHandlerImpl
        """
        return StatusSocketHandlerImpl(balance, *args, **kwargs)

    return StatusSocketHandler


class ThreadingSocketServer(ThreadingMixIn, UnixStreamServer):
    """
    Unix socket server with threading
    """
    pass


class StatusSocket(BaseStatusSocket):
    """
    Create a Unix domain socket which emits a summary of the Onionbalance
    status when a client connects.
    """

    def __init__(self, status_socket_location, balance: Onionbalance):
        """
        Create the Unix domain socket status server and start in a thread

        Example::
            socat - unix-connect:/var/run/onionbalance/control

            {"services": [{"instances": [{"introModified": "2020-06-16 19:35:17", "ipsNum": 3, "onionAddress": "vkmiy6biqcyphtx5exswxl5sjus2vn2b6pzir7lz5akudhwbqk5muead.onion"}], "onionAddress": "bvy46sg2b5dokczabwv2pabqlrps3lppweyrebhat6gjieo2avojdvad.onion.onion", "timestamp": "2020-06-16 19:36:01"}]}
        """
        super().__init__(status_socket_location)

        self.cleanup_socket_file()

        logger.debug("Creating status socket at %s", self.unix_socket_filename)
        try:
            self.server = ThreadingSocketServer(self.unix_socket_filename,
                                                create_status_socket_handler(balance))

            # Start running the socket server in a another thread
            server_thread = threading.Thread(target=self.server.serve_forever)
            server_thread.daemon = True  # Exit daemon when main thread stops
            server_thread.start()

        except (OSError, socket.error):
            logger.error("Could not start status socket at %s. Does the path "
                         "exist? Do you have permission?",
                         status_socket_location)



# -*- coding: utf-8 -*-
import os

"""
Define default config options for the management server
"""

# Set default configuration options for the management server

REPLICAS = 2
HSDIR_SET = 3  # Publish each descriptor to 3 consecutive HSDirs
MAX_INTRO_POINTS = 10
DESCRIPTOR_OVERLAP_PERIOD = 60 * 60
DESCRIPTOR_UPLOAD_PERIOD = 60 * 60  # Re-upload descriptor every hour
REFRESH_INTERVAL = 10 * 60
PUBLISH_CHECK_INTERVAL = 5 * 60
INITIAL_DELAY = 45  # Wait for instance descriptors before publishing

LOG_LOCATION = os.environ.get('ONIONBALANCE_LOG_LOCATION')
LOG_LEVEL = os.environ.get('ONIONBALANCE_LOG_LEVEL', 'info')

STATUS_SOCKET_LOCATION = os.environ.get('ONIONBALANCE_STATUS_SOCKET_LOCATION',
                                        '/var/run/onionbalance/control')
TOR_CONTROL_PASSWORD = None

# Upload multiple distinct descriptors containing different subsets of
# the available introduction points
DISTINCT_DESCRIPTORS = True

# Store global data about onion services and their instance nodes.
services = []

controller = None



import os

# Parameters definining Onionbalance behavior

# How long to wait for onionbalance to bootstrap before starting periodic
# events (in seconds)
INITIAL_CALLBACK_DELAY = 45

# Every how often we should be fetching instance descriptors (in seconds)
FETCH_DESCRIPTOR_FREQUENCY = 10 * 60
FETCH_DESCRIPTOR_FREQUENCY_TESTNET = 20

# Every how often we should be checking whether we should publish our frontend
# descriptor (in seconds). Triggering this callback doesn't mean we will
# actually upload a descriptor. We only upload a descriptor if it has expired,
# the intro points have changed, etc.
PUBLISH_DESCRIPTOR_CHECK_FREQUENCY = 5 * 60
PUBLISH_DESCRIPTOR_CHECK_FREQUENCY_TESTNET = 10

# How long should we keep a frontend descriptor before we expire it (in
# seconds)?
FRONTEND_DESCRIPTOR_LIFETIME = 60 * 60
FRONTEND_DESCRIPTOR_LIFETIME_TESTNET = 20

# How many intros should we use from each instance in the final frontend
# descriptor?
# [TODO: This makes no attempt to hide the use of onionbalance. In the future we
# should be smarter and sneakier here.]
N_INTROS_PER_INSTANCE = 2

# If we last received a descriptor for this instance more than
# INSTANCE_DESCRIPTOR_TOO_OLD seconds ago, consider the instance to be down.
INSTANCE_DESCRIPTOR_TOO_OLD = 60 * 60
INSTANCE_DESCRIPTOR_TOO_OLD_TESTNET = 20

# Parameters defined by HSv3 spec and little-t-tor implementation

# Number of replicas per descriptor
HSDIR_N_REPLICAS = 2
# How many uploads per replica
# [TODO: Get these from the consensus instead of hardcoded]
HSDIR_SPREAD_STORE = 4

# Max descriptor size (in bytes) (see hs_cache_get_max_descriptor_size() in
# little-t-tor)
MAX_DESCRIPTOR_SIZE = 50000

# Misc parameters

DEFAULT_LOG_LEVEL = os.environ.get('ONIONBALANCE_LOG_LEVEL', 'warning')



#!/usr/bin/python
# Copyright 2014-2019, The Tor Project, Inc
# See LICENSE for licensing information

"""
   Reference implementations for the ed25519 tweaks that Tor uses.

   Includes self-tester and test vector generator.
"""

from . import slow_ed25519
from .slow_ed25519 import *

import os
import random
import unittest
import binascii
import textwrap

#define a synonym that doesn't look like 1
ell = l

# This replaces expmod above and makes it go a lot faster.
slow_ed25519.expmod = pow

def blindESK(esk, param):
    mult = 2**(b-2) + sum(2**i * bit(param,i) for i in range(3,b-2))
    s = decodeint(esk[:32])
    s_prime = (s * mult) % ell
    k = esk[32:]
    assert(len(k) == 32)
    k_prime = H(b"Derive temporary signing key hash input" + k)[:32]
    return encodeint(s_prime) + k_prime

def blindPK(pk, param):
    mult = 2**(b-2) + sum(2**i * bit(param,i) for i in range(3,b-2))
    P = decodepoint(pk)
    return encodepoint(scalarmult(P, mult))

def expandSK(sk):
    h = H(sk)
    a = 2**(b-2) + sum(2**i * bit(h,i) for i in range(3,b-2))
    k = b''.join([bytes([h[i]]) for i in range(b//8,b//4)])
    assert len(k) == 32
    return encodeint(a)+k

def publickeyFromESK(h):
    a = decodeint(h[:32])
    A = scalarmult(B,a)
    return encodepoint(A)

def signatureWithESK(m,h,pk):
    a = decodeint(h[:32])
    r = Hint(b''.join([bytes([h[i]]) for i in range(b//8,b//4)]) + m)
    R = scalarmult(B,r)
    S = (r + Hint(encodepoint(R) + pk + m) * a) % l
    return encodepoint(R) + encodeint(S)

def newSK():
    return os.urandom(32)

def random_scalar(entropy_f): # 0..L-1 inclusive
    # reduce the bias to a safe level by generating 256 extra bits
    oversized = int(binascii.hexlify(entropy_f(32+32)), 16)
    return oversized % ell

# ------------------------------------------------------------

MSG = b"This is extremely silly. But it is also incredibly serious business!"

class SelfTest(unittest.TestCase):

    def _testSignatures(self, esk, pk):
        sig = signatureWithESK(MSG, esk, pk)
        checkvalid(sig, MSG, pk)
        bad = False
        try:
            checkvalid(sig, MSG*2, pk)
            bad = True
        except Exception:
            pass

        self.assertFalse(bad)

    def testExpand(self):
        sk = newSK()
        pk = publickey(sk)
        esk = expandSK(sk)
        sig1 = signature(MSG, sk, pk)
        sig2 = signatureWithESK(MSG, esk, pk)
        self.assertEqual(sig1, sig2)

    def testSignatures(self):
        sk = newSK()
        esk = expandSK(sk)
        pk = publickeyFromESK(esk)
        pk2 = publickey(sk)
        self.assertEqual(pk, pk2)

        self._testSignatures(esk, pk)

    def testBlinding(self):
        sk = newSK()
        esk = expandSK(sk)
        pk = publickeyFromESK(esk)
        param = os.urandom(32)
        besk = blindESK(esk, param)
        bpk = blindPK(pk, param)
        bpk2 = publickeyFromESK(besk)
        self.assertEqual(bpk, bpk2)

        self._testSignatures(besk, bpk)

    def testIdentity(self):
        # Base point:
        # B is the unique point (x, 4/5) \in E for which x is positive
        By = 4 * inv(5)
        Bx = xrecover(By)
        B = [Bx % q,By % q]

        # Get identity E by doing: E = l*B, where l is the group order
        identity = scalarmult(B, ell)

        # Get identity E by doing: E = l*A, where A is a random point
        sk = newSK()
        pk = decodepoint(publickey(sk))
        identity2 = scalarmult(pk, ell)

        # Check that identities match
        assert(identity == identity2)
        # Check that identity is the point (0,1)
        assert(identity == [0,1])

        # Check identity element: a*E = E, where a is a random scalar
        scalar = random_scalar(os.urandom)
        result = scalarmult(identity, scalar)
        assert(result == identity == identity2)

# ------------------------------------------------------------

# From pprint.pprint([ binascii.b2a_hex(os.urandom(32)) for _ in xrange(8) ])
RAND_INPUTS = [
  '26c76712d89d906e6672dafa614c42e5cb1caac8c6568e4d2493087db51f0d36',
  'fba7a5366b5cb98c2667a18783f5cf8f4f8d1a2ce939ad22a6e685edde85128d',
  '67e3aa7a14fac8445d15e45e38a523481a69ae35513c9e4143eb1c2196729a0e',
  'd51385942033a76dc17f089a59e6a5a7fe80d9c526ae8ddd8c3a506b99d3d0a6',
  '5c8eac469bb3f1b85bc7cd893f52dc42a9ab66f1b02b5ce6a68e9b175d3bb433',
  'eda433d483059b6d1ff8b7cfbd0fe406bfb23722c8f3c8252629284573b61b86',
  '4377c40431c30883c5fbd9bc92ae48d1ed8a47b81d13806beac5351739b5533d',
  'c6bbcce615839756aed2cc78b1de13884dd3618f48367a17597a16c1cd7a290b']

# From pprint.pprint([ binascii.b2a_hex(os.urandom(32)) for _ in xrange(8) ])
BLINDING_PARAMS = [
  '54a513898b471d1d448a2f3c55c1de2c0ef718c447b04497eeb999ed32027823',
  '831e9b5325b5d31b7ae6197e9c7a7baf2ec361e08248bce055908971047a2347',
  'ac78a1d46faf3bfbbdc5af5f053dc6dc9023ed78236bec1760dadfd0b2603760',
  'f9c84dc0ac31571507993df94da1b3d28684a12ad14e67d0a068aba5c53019fc',
  'b1fe79d1dec9bc108df69f6612c72812755751f21ecc5af99663b30be8b9081f',
  '81f1512b63ab5fb5c1711a4ec83d379c420574aedffa8c3368e1c3989a3a0084',
  '97f45142597c473a4b0e9a12d64561133ad9e1155fe5a9807fe6af8a93557818',
  '3f44f6a5a92cde816635dfc12ade70539871078d2ff097278be2a555c9859cd0']

PREFIX = "ED25519_"

def writeArray(name, array):
    print("static const char *{prefix}{name}[] = {{".format(
        prefix=PREFIX,name=name))
    for a in array:
        h = a.hex()
        if len(h) > 70:
            h1 = h[:70]
            h2 = h[70:]
            print('  "{0}"\n      "{1}",'.format(h1,h2))
        else:
            print('  "{0}",'.format(h))
    print("};\n")

def comment(text, initial="/**"):
    print(initial)
    print(textwrap.fill(text,initial_indent=" * ",subsequent_indent=" * "))
    print(" */")

def makeTestVectors():
    comment("""Test vectors for our ed25519 implementation and related
               functions. These were automatically generated by the
               ed25519_exts_ref.py script.""", initial="/*")


    comment("""Secret key seeds used as inputs for the ed25519 test vectors.
               Randomly generated. """)
    secretKeys = [ bytes.fromhex(r) for r in RAND_INPUTS ]
    writeArray("SECRET_KEYS", secretKeys)

    comment("""Secret ed25519 keys after expansion from seeds. This is how Tor
               represents them internally.""")
    expandedSecretKeys = [ expandSK(sk) for sk in secretKeys ]
    writeArray("EXPANDED_SECRET_KEYS", expandedSecretKeys)

    comment("""Public keys derived from the above secret keys""")
    publicKeys = [ publickey(sk) for sk in secretKeys ]
    writeArray("PUBLIC_KEYS", publicKeys)

    comment("""Parameters used for key blinding tests. Randomly generated.""")
    blindingParams =  [ binascii.a2b_hex(r) for r in BLINDING_PARAMS ]
    writeArray("BLINDING_PARAMS", blindingParams)

    comment("""Blinded secret keys for testing key blinding.  The nth blinded
               key corresponds to the nth secret key blidned with the nth
               blinding parameter.""")
    writeArray("BLINDED_SECRET_KEYS",
               (blindESK(expandSK(sk), bp)
                for sk,bp in zip(secretKeys,blindingParams)))

    comment("""Blinded public keys for testing key blinding.  The nth blinded
               key corresponds to the nth public key blidned with the nth
               blinding parameter.""")
    writeArray("BLINDED_PUBLIC_KEYS",
               (blindPK(pk, bp) for pk,bp in zip(publicKeys,blindingParams)))

    comment("""Signatures of the public keys, made with their corresponding
               secret keys.""")
    writeArray("SELF_SIGNATURES",
               (signature(pk, sk, pk) for pk,sk in zip(publicKeys,secretKeys)))



if __name__ == '__main__':
    import sys
    if len(sys.argv) == 1 or sys.argv[1] not in ("SelfTest", "MakeVectors"):
        print ("You should specify one of 'SelfTest' or 'MakeVectors'")
        sys.exit(1)
    if sys.argv[1] == 'SelfTest':
        unittest.main()
    else:
        makeTestVectors()
